// Copyright (C) 2016 EDF
// All Rights Reserved
// This code is published under the GNU Lesser General Public License (GNU LGPL)
#define BOOST_TEST_MODULE testLocalLinearRegression
#define BOOST_TEST_DYN_LINK
#include <memory>
#include <functional>
#include <boost/test/unit_test.hpp>
#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>
#include <boost/random/uniform_01.hpp>
#include <Eigen/Dense>
#include "geners/BinaryFileArchive.hh"
#include "geners/Record.hh"
#include "geners/Reference.hh"
#include "StOpt/core/utils/constant.h"
#include "StOpt/core/grids/RegularSpaceGridGeners.h"
#include "StOpt/core/grids/LinearInterpolator.h"
#include "StOpt/regression/LocalLinearRegressionGeners.h"
#include "StOpt/regression/ContinuationValueGeners.h"
#include "StOpt/regression/ContinuationCutsGeners.h"
#include "StOpt/regression/ContinuationCutsGridAdaptNonConcaveGeners.h"

using namespace std;
using namespace Eigen;
using namespace gs;
using namespace StOpt;


/// For Clang < 3.7 (and above ?) to be compatible GCC 5.1 and above
namespace boost
{
namespace unit_test
{
namespace ut_detail
{
std::string normalize_test_case_name(const_string name)
{
    return (name[0] == '&' ? std::string(name.begin() + 1, name.size() - 1) : std::string(name.begin(), name.size()));
}
}
}
}

double accuracyEqual = 1e-7;
double accuracyNearEqual = 0.5;

#if defined   __linux
#include <fenv.h>
#define enable_abort_on_floating_point_exception() feenableexcept(FE_DIVBYZERO | FE_INVALID)
#endif

// utilities developed because VS2010 doesn't support auto
class Func1
{
public :
    Func1() {}

    double operator()(const double &p_x) const
    {
        return (1 + 0.1 * p_x) * (1 + 0.1 * p_x) + p_x + 2.;
    }
};

class Func2
{

public :

    Func2() {}

    double operator()(const int p_mem, const double &p_x) const
    {
        return  p_x * p_x + p_x * (p_mem + 1) + 1.;
    }
};

class  Func3
{

public :

    Func3() {}

    double operator()(const double &p_x, const double &p_y) const
    {
        return  p_x * p_y + p_x + p_y + 1.;
    }
};

// test for different  dimensions
void testDimension(const bool &p_bRotate, const int &p_nDim, const int &p_nbSimul, const int &p_nMesh, const double &p_accuracyRegression)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif
    // create the mesh
    ArrayXi  nbMesh = ArrayXi::Constant(p_nDim, p_nMesh);
    // generator Mersene Twister
    boost::mt19937 generator;
    boost::normal_distribution<double> alea_n;
    boost::variate_generator<boost::mt19937 &, boost::normal_distribution<double> > normal_random(generator, alea_n) ;
    std::function< double(double) > f = Func1();
    ArrayXXd x = ArrayXXd::Random(p_nDim, p_nbSimul);
    // second member to regress
    ArrayXd toRegress(p_nbSimul);
    ArrayXd toReal(p_nbSimul);
    for (int is  = 0;  is < p_nbSimul; ++is)
    {
        double prod = f(x(0, is));
        for (int id = 1 ; id < p_nDim ; ++id)
            prod *=  f(x(id, is));
        toRegress(is) = prod +  4 * normal_random();
        toReal(is) = prod;
    }
    // check if regression in t=0
    {
        // constructor if time t  = 0
        bool bZeroDate = true;
        // constructor
        LocalLinearRegression localRegressor(nbMesh, p_bRotate);
        localRegressor.updateSimulations(bZeroDate, x);
        ArrayXd  regressedValues = localRegressor.getAllSimulations(toRegress);

        double average = toRegress.mean();
        for (int is = 0; is < p_nbSimul; ++is)
        {
            if ((fabs(regressedValues(is)) > tiny) && (fabs(average) > tiny))
                BOOST_CHECK(fabs(regressedValues(is) - average) <  accuracyEqual * max(fabs(average), fabs(regressedValues(is))));
        }
    }

    {
        // constructor not zero
        bool bZeroDate = 0;

        // constructor
        LocalLinearRegression localRegressor(nbMesh, p_bRotate);
        localRegressor.updateSimulations(bZeroDate, x);
        // first regression
        ArrayXd  regressedValues = localRegressor.getAllSimulations(toRegress);
        // then just calculate function basis coefficient
        ArrayXd regressedFuntionCoeff = localRegressor.getCoordBasisFunction(toRegress);
        for (int is = 0; is < p_nbSimul; ++is)
        {
            Map<ArrayXd> xloc(x.col(is).data(), p_nDim);
            BOOST_CHECK_CLOSE(regressedValues(is), localRegressor.getValue(xloc, regressedFuntionCoeff), accuracyNearEqual);
        }
        // Check that regression is correct
        double erMax = 0. ;
        for (int is = 0; is < p_nbSimul; ++is)
            erMax = std::max(erMax, std::fabs((regressedValues(is) - toReal(is)) / toReal(is)));
        // now get back all
        BOOST_CHECK_CLOSE(1. + erMax, 1., p_accuracyRegression);
        // get back all values once for all
        ArrayXd regressedAllValues  = localRegressor.getValues(x, regressedFuntionCoeff) ;
        for (int is = 0; is < p_nbSimul; ++is)
        {
            BOOST_CHECK_CLOSE(regressedValues(is), regressedAllValues(is), accuracyNearEqual);
        }

    }
}


// 1D check single second member regression
BOOST_AUTO_TEST_CASE(testLocalRegression1D)
{

#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif
    // dimension
    int nDim = 1 ;
    // Number of simulations and parameters for regression.
    int nMesh = 16 ;
    ArrayXi  nbMesh = ArrayXi::Constant(nDim, nMesh);
    int nbSimul = 2000000;
    // accuracy
    double accuracy = 5; // percentage error
    testDimension(false, nDim, nbSimul, nMesh, accuracy);
}

// 2D check single second member
BOOST_AUTO_TEST_CASE(testLocalRegression2D)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif
    // dimension
    int nDim = 2 ;
    // Number of simulations and parameters for regression.
    int nMesh = 6 ;
    ArrayXi  nbMesh = ArrayXi::Constant(nDim, nMesh);
    int nbSimul = 2000000;
    // accuracy
    double accuracy = 5; // percentage error
    testDimension(false, nDim, nbSimul, nMesh, accuracy);
    // compiler bug on mips
#if !defined(mips) && !defined(__mips__) && !defined(__mips)
    testDimension(true, nDim, nbSimul, nMesh, accuracy);
#endif
}

// 3D check single second member
BOOST_AUTO_TEST_CASE(testLocalRegression3D)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif
    // dimension
    int nDim = 3 ;
    // Number of simulations and parameters for regression.
    // int nMesh =40 ;
    int nMesh = 8 ;
    ArrayXi  nbMesh = ArrayXi::Constant(nDim, nMesh);
    int nbSimul = 2000000;
    // accuracy
    double accuracy = 10; // percentage error
    testDimension(false, nDim, nbSimul, nMesh, accuracy);
}

// 1D check multiple second members
BOOST_AUTO_TEST_CASE(testLocalRegression1DMultipleSecondMember)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif
    // dimension
    int nDim = 1 ;
    // number of second member
    int nMember = 5 ;
    // Number of simulations and parameters for regression.
    int nMesh = 4 ;
    ArrayXi  nbMesh = ArrayXi::Constant(nDim, nMesh);
    int nbSimul = 20000;
    // generator Mersene Twister
    boost::mt19937 generator;
    boost::normal_distribution<double> alea_n;
    boost::variate_generator<boost::mt19937 &, boost::normal_distribution<double> > normal_random(generator, alea_n) ;
    std::function< double(int, double) >  f = Func2();
    ArrayXXd x = ArrayXXd::Random(nDim, nbSimul);
    // second member to regress
    ArrayXXd toRegress(nMember, nbSimul);
    for (int imem = 0; imem < nMember; ++imem)
        for (int is  = 0;  is < nbSimul; ++is)
            toRegress(imem, is) = f(imem, x(0, is)) +  normal_random();

    // check if regression in t=0
    {
        // constructor if time t  = 0
        bool bZeroDate = true;
        // constructor
        LocalLinearRegression localRegressor(nbMesh);
        localRegressor.updateSimulations(bZeroDate, x);
        // regression
        ArrayXXd  regressedValues = localRegressor.getAllSimulationsMultiple(toRegress);

        for (int imem = 0; imem < nMember; ++imem)
        {
            double average = toRegress.row(imem).mean();
            for (int is = 0; is < nbSimul; ++is)
                BOOST_CHECK_CLOSE(regressedValues(imem, is), average, accuracyEqual);
        }
    }
    // t is not zero
    {
        // constructor not zero
        bool bZeroDate = 0;
        // constructor
        LocalLinearRegression localRegressor(bZeroDate, x, nbMesh);
        // regression
        ArrayXXd  regressedValues = localRegressor.getAllSimulationsMultiple(toRegress);
        // regression function coefficient
        ArrayXXd regressedFuntionCoeff = localRegressor.getCoordBasisFunctionMultiple(toRegress);

        // now regress all second member and test
        for (int imem = 0 ; imem <  nMember; ++imem)
        {
            ArrayXd toRegress1D(nbSimul) ;
            toRegress1D = toRegress.row(imem);
            ArrayXd regressedValues1D = localRegressor.getAllSimulations(toRegress1D);
            for (int is = 0; is < nbSimul; ++is)
                BOOST_CHECK_CLOSE(regressedValues(imem, is), regressedValues1D(is), accuracyEqual);
            ArrayXd regressedFuntionCoeff1D = localRegressor.getCoordBasisFunction(toRegress1D);
            for (int im = 0; im < regressedFuntionCoeff1D.size(); ++im)
                BOOST_CHECK_CLOSE(regressedFuntionCoeff(imem, im), regressedFuntionCoeff1D(im), accuracyEqual);
        }
    }
}




// test serialization LocalLinearRegression
void  TestLocalRegressionSerialization(const int &p_nDim, const int &p_nbSimul, const int &p_nbMesh)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif
    ArrayXi  nbMesh = ArrayXi::Constant(p_nDim, p_nbMesh);
    // generator Mersene Twister
    boost::mt19937 generator;
    boost::normal_distribution<double> alea_n;
    boost::variate_generator<boost::mt19937 &, boost::normal_distribution<double> > normal_random(generator, alea_n) ;
    std::function< double(double) > f = Func1();
    ArrayXXd x = ArrayXXd::Random(p_nDim, p_nbSimul);
    // to store basis function
    ArrayXd basisFunction, regressedValues	;
    // test archive
    {
        // second member to regress with one stock
        ArrayXd toRegress(p_nbSimul);
        for (int is  = 0;  is < p_nbSimul; ++is)
        {
            double prod = f(x(0, is));
            for (int id = 1 ; id < p_nDim ; ++id)
                prod *=  f(x(id, is));
            double uncertainty = 4 * normal_random();
            toRegress(is) = prod +  uncertainty;
        }
        // conditional expectation
        shared_ptr<LocalLinearRegression> localRegressor = make_shared<LocalLinearRegression>(false, x, nbMesh);

        // regress  directly with regressor
        regressedValues = localRegressor->getAllSimulations(toRegress);

        // basis functions
        basisFunction =  localRegressor->getCoordBasisFunction(toRegress);

        // test
        for (int is  = 0; is < p_nbSimul; ++is)
        {
            double regVal = localRegressor->reconstructionASim(is, basisFunction);
            BOOST_CHECK_CLOSE(regressedValues(is), regVal, accuracyEqual);
        }

        // archive
        {
            BinaryFileArchive ar("archiveLLR", "w");
            ar << Record(*localRegressor, "Regressor", "Top") ;
        }
        {
            BinaryFileArchive ar("archiveLLR", "r");
            LocalLinearRegression reg;
            Reference< LocalLinearRegression> (ar, "Regressor", "Top").restore(0, &reg);
            for (int is = 0; is < x.size(); ++is)
                BOOST_CHECK_CLOSE(regressedValues(is), reg.getValue(x.col(is), basisFunction), accuracyEqual);

        }
    }
}

BOOST_AUTO_TEST_CASE(testLocalRegressionSerialization)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif
    TestLocalRegressionSerialization(1, 10, 1);
}

/// test serialization for continuation values
void TestContinuationValue(const int &p_nDim, const int &p_nbSimul, const int &p_nbMesh)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif
    // nb stock points
    int sizeForStock  = 4;
    ArrayXi  nbMesh = ArrayXi::Constant(p_nDim, p_nbMesh);
    // generator Mersene Twister
    boost::mt19937 generator;
    boost::normal_distribution<double> alea_n;
    boost::variate_generator<boost::mt19937 &, boost::normal_distribution<double> > normal_random(generator, alea_n) ;
    std::function< double(double) > f = Func1();
    ArrayXXd x = ArrayXXd::Random(p_nDim, p_nbSimul);     // test archive
    ArrayXXd  regressedValues;
    {

        // second member to regress with one stock
        ArrayXXd toRegress(sizeForStock, p_nbSimul);
        for (int is  = 0;  is < p_nbSimul; ++is)
        {
            double prod = f(x(0, is));
            for (int id = 1 ; id < p_nDim ; ++id)
                prod *=  f(x(id, is));
            double uncertainty = 4 * normal_random();
            for (int j = 0; j < sizeForStock; ++j)
            {
                toRegress(j, is) = prod * (j + 1) +  uncertainty;
            }
        }
        // grid for stock
        Eigen::ArrayXd lowValues(1), step(1);
        lowValues(0) = 0. ;
        step(0) = 1;
        Eigen::ArrayXi  nbStep(1);
        nbStep(0) = sizeForStock - 1;
        // grid
        shared_ptr< RegularSpaceGrid > regular = make_shared<RegularSpaceGrid>(lowValues, step, nbStep);
        // conditional expectation
        shared_ptr<LocalLinearRegression> localRegressor = make_shared<LocalLinearRegression>(false, x, nbMesh);

        // regress  directly with regressor
        regressedValues = localRegressor->getAllSimulationsMultiple(toRegress);
        // creation continuation value object
        ContinuationValue  continuation(regular, localRegressor,  toRegress.transpose());

        // regress with continuation value object
        ArrayXd ptStock(1) ;
        ptStock(0) = sizeForStock / 2;
        ArrayXd regressedByContinuation = continuation.getAllSimulations(ptStock);
        for (int is  = 0;  is < p_nbSimul; ++is)
            BOOST_CHECK_CLOSE(regressedByContinuation(is), regressedValues(sizeForStock / 2, is), accuracyEqual);

        // default non compression
        BinaryFileArchive ar("archiveLLR1", "w");
        ar << Record(continuation, "FirstContinuation", "Top") ;

        // iterate on grid points
        // back grid iterator from continuation value object
        shared_ptr<GridIterator> iterRegGrid  = continuation.getGrid()->getGridIterator();
        // get back the condition expectation operator
        std::shared_ptr<BaseRegression > condExp = continuation.getCondExp();
        // get back basis functions for all grid points
        const Eigen::ArrayXXd &basisFuncValues = continuation.getValues();
        // stock identifier
        int iStock = 0;
        while (iterRegGrid->isValid())
        {
            // first version using the stock level  so with interpolation
            ArrayXd pointCoord = iterRegGrid->getCoordinate();
            ArrayXd regressedByContinuation1 = continuation.getAllSimulations(pointCoord);
            // second method without interpolation : get back directly the basis function associated to the stock identifier
            ArrayXd regressedByContinuation2 = condExp->reconstruction(basisFuncValues.col(iStock));
            // third method to be used in simulation
            ArrayXd regressedByContinuation3(p_nbSimul);
            for (int is  = 0;  is < p_nbSimul; ++is)
                regressedByContinuation3(is) =  condExp->getValue(x.col(is), basisFuncValues.col(iStock));
            for (int is  = 0;  is < p_nbSimul; ++is)
            {
                BOOST_CHECK_CLOSE(regressedByContinuation1(is), regressedByContinuation2(is), accuracyEqual);
                BOOST_CHECK_CLOSE(regressedByContinuation1(is), regressedByContinuation3(is), accuracyEqual);
            }
            iterRegGrid->next();
            iStock += 1;
        }
    }
    {
        // read archive
        BinaryFileArchive ar("archiveLLR1", "r");
        ContinuationValue contRead;
        Reference< ContinuationValue >(ar, "FirstContinuation", "Top").restore(0, &contRead);
        ArrayXd ptStock(1) ;
        ptStock(0) = sizeForStock / 2;
        for (int is  = 0;  is < p_nbSimul; ++is)
            BOOST_CHECK_CLOSE(contRead.getValue(ptStock, x.col(is)), regressedValues(sizeForStock / 2, is), accuracyEqual);
    }
}



BOOST_AUTO_TEST_CASE(testContinuation)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif

    TestContinuationValue(1, 10, 1);
}

// test serialization for continuation cuts values
void TestContinuationCuts(const int &p_nDim, const int &p_nbSimul, const int &p_nbMesh)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif
    // nb stock points
    int sizeForStock  = 4;
    ArrayXi  nbMesh = ArrayXi::Constant(p_nDim, p_nbMesh);
    // generator Mersene Twister
    boost::mt19937 generator;
    boost::normal_distribution<double> alea_n;
    boost::variate_generator<boost::mt19937 &, boost::normal_distribution<double> > normal_random(generator, alea_n) ;
    std::function< double(double) > f = Func1();
    ArrayXXd x = ArrayXXd::Random(p_nDim, p_nbSimul);     // test archive
    ArrayXXd  regressedValues;
    ArrayXXd  hyperStock(1, 2) ;

    // grid for storage
    Eigen::ArrayXd lowValues(1), step(1);
    lowValues(0) = 0. ;
    step(0) = 1; // step for storage equal to 1
    Eigen::ArrayXi  nbStep(1);
    nbStep(0) = sizeForStock - 1;
    // grid
    shared_ptr< RegularSpaceGrid > regular = make_shared<RegularSpaceGrid>(lowValues, step, nbStep);


    {
        // second member to regress with one stock  (one storage, so one sensibility)
        ArrayXXd toRegress(sizeForStock, p_nbSimul);
        ArrayXXd toRegressCut(sizeForStock, p_nbSimul * 2);
        for (int is  = 0;  is < p_nbSimul; ++is)
        {
            double prod = f(x(0, is));
            for (int id = 1 ; id < p_nDim ; ++id)
                prod *=  f(x(id, is));
            double uncertainty = 4 * normal_random();
            for (int j = 0; j < sizeForStock; ++j)
            {
                toRegress(j, is) = prod * (j + 1) +  uncertainty;
                toRegressCut(j, is) = prod * (j + 1) +  uncertainty;
                toRegressCut(j, is + p_nbSimul) = prod ; // sensibility equal to prod
            }
        }
        // conditional expectation
        shared_ptr<LocalLinearRegression> localRegressor = make_shared<LocalLinearRegression>(false, x, nbMesh);

        // regress  directly with regressor
        regressedValues = localRegressor->getAllSimulationsMultiple(toRegress);

        // creation continuation value object
        ContinuationCuts  contCut(regular, localRegressor,  toRegressCut.transpose());

        // calculate all cuts
        hyperStock(0, 0) = lowValues(0);
        hyperStock(0, 1) = lowValues(0) + nbStep(0) * step(0);

        ArrayXXd regressedCuts = contCut.getCutsAllSimulations(hyperStock);

        // test only values
        shared_ptr<GridIterator> iterGrid = regular->getGridIterator();
        while (iterGrid->isValid())
        {
            for (int is  = 0;  is < p_nbSimul; ++is)
            {
                // coordinates
                ArrayXd pointCoord = iterGrid->getCoordinate();
                // reconstruct value from cuts coefficients
                double valReconst = regressedCuts(is, iterGrid->getCount()) + regressedCuts(is + p_nbSimul, iterGrid->getCount()) * pointCoord(0);
                BOOST_CHECK_CLOSE(valReconst, regressedValues(iterGrid->getCount(), is), accuracyEqual);
            }
            iterGrid->next();
        }

        // default non compression
        BinaryFileArchive ar("archiveLLR2", "w");
        ar << Record(contCut, "FirstContinuation", "Top") ;

    }
    {
        // read archive
        BinaryFileArchive ar("archiveLLR2", "r");
        ContinuationCuts contCutRead;
        Reference< ContinuationCuts >(ar, "FirstContinuation", "Top").restore(0, &contCutRead);

        // test only values
        shared_ptr<GridIterator> iterGrid = regular->getGridIterator();
        while (iterGrid->isValid())
        {
            for (int is  = 0;  is < p_nbSimul; ++is)
            {
                // coordinates
                ArrayXd pointCoord = iterGrid->getCoordinate();

                // cut from a sim
                ArrayXXd regCuts = contCutRead.getCutsASim(hyperStock, x.col(is));

                // reconstruct value from cuts coefficients
                double valReconst = regCuts(0, iterGrid->getCount()) + regCuts(1, iterGrid->getCount()) * pointCoord(0);
                BOOST_CHECK_CLOSE(valReconst, regressedValues(iterGrid->getCount(), is), accuracyEqual);
            }
            iterGrid->next();
        }
    }
}



BOOST_AUTO_TEST_CASE(testContCut)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif

    TestContinuationCuts(1, 10, 1);
}


// test serialization for continuation cuts values for 2D adapted grids
void TestContinuationCutsGridAdapt2D(const int &p_nDim, const int &p_nbSimul, const int &p_nbMesh)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif
    ArrayXi  nbMesh = ArrayXi::Constant(p_nDim, p_nbMesh);
    // generator Mersene Twister
    boost::mt19937 generator;
    boost::normal_distribution<double> alea_n;
    boost::variate_generator<boost::mt19937 &, boost::normal_distribution<double> > normal_random(generator, alea_n) ;
    std::function< double(double) > f = Func1();
    ArrayXXd x = ArrayXXd::Random(p_nDim, p_nbSimul);     // test archive
    ArrayXXd  regressedValues;
    ArrayXXd  hyperStock(2, 2) ;

    // grid for storage
    double xMin = 0;
    double xMax = 5.;
    int nbMeshX = 5;
    double yMin = 0.;
    double yMax = 4.;
    int nbMeshY = 4 ;

    // nb stock points
    int sizeForStock  = (nbMeshX + 1) * (nbMeshY + 1);

    // grid
    shared_ptr< GridAdapt2D > grid2D = make_shared<GridAdapt2D>(xMin, xMax, nbMeshX, yMin, yMax, nbMeshY);

    {
        // second member to regress with one stock  (one storage, so one sensibility)
        ArrayXXd toRegress(sizeForStock, p_nbSimul);
        ArrayXXd toRegressCut(sizeForStock, p_nbSimul * 3);
        for (int is  = 0;  is < p_nbSimul; ++is)
        {
            double prod = f(x(0, is));
            for (int id = 1 ; id < p_nDim ; ++id)
                prod *=  f(x(id, is));
            double uncertainty = 4 * normal_random();
            for (int j = 0; j < nbMeshY + 1; ++j)
                for (int i = 0; i < nbMeshX + 1; ++i)
                {
                    int istock = i + j * (nbMeshX + 1);
                    toRegress(istock, is) = prod * istock  +  uncertainty;
                    toRegressCut(istock, is) = prod * istock +  uncertainty;
                    toRegressCut(istock, is + p_nbSimul) = prod ; // sensibility equal to prod in x dimension
                    toRegressCut(istock, is + 2 * p_nbSimul) = prod ; // sensibility equal to prod in y dimension
                }

        }
        // conditional expectation
        shared_ptr<LocalLinearRegression> localRegressor = make_shared<LocalLinearRegression>(false, x, nbMesh);

        // regress  directly with regressor
        regressedValues = localRegressor->getAllSimulationsMultiple(toRegress);

        // creation continuation value object
        ContinuationCutsGridAdaptNonConcave  contCut(grid2D, localRegressor,  toRegressCut.transpose());

        // calculate all cuts
        hyperStock(0, 0) = xMin;
        hyperStock(0, 1) = xMax;
        hyperStock(1, 0) = yMin;
        hyperStock(1, 1) = yMax;

        vector< pair<shared_ptr< GridAdaptBase>, shared_ptr<ArrayXXd>  > >   regressedCutsPMesh = contCut.getCutsAllSimulations(hyperStock);

        // test only values
        ArrayXi   pointNumber(4);
        ArrayXd pointCoord(2);
        for (const auto &pairMA : regressedCutsPMesh)
        {
            // for all points on the sub grid
            std::vector< Eigen::ArrayXd > points = pairMA.first->getPoints();

            for (size_t i = 0; i < points.size(); ++i)
            {
                pointCoord(0) = points[i][0];
                pointCoord(1) = points[i][1];
                // locate position in the initial grid
                int ipos = static_cast<int>(pointCoord(0)) +  static_cast<int>(pointCoord(1)) * (nbMeshX + 1);
                for (int isim = 0; isim < p_nbSimul; ++isim)
                {
                    // reconstruct value from cuts coefficients
                    double valReconst = (*pairMA.second)(isim, i) + (*pairMA.second)(isim + p_nbSimul, i) * pointCoord(0) + (*pairMA.second)(isim + 2 * p_nbSimul, i) * pointCoord(1);
                    BOOST_CHECK_CLOSE(valReconst, regressedValues(ipos, isim), accuracyEqual);
                }
            }
        }
        for (int isim = 0; isim < p_nbSimul;  ++isim)
        {
            vector< pair<shared_ptr< GridAdaptBase>, shared_ptr<ArrayXXd>  > >   regressedCutsPMesh = contCut.getCutsASim(hyperStock, isim);

            for (const auto &pairMA : regressedCutsPMesh)
            {
                // for all points on the sub grid
                std::vector< Eigen::ArrayXd > points = pairMA.first->getPoints();

                for (size_t i = 0; i < points.size(); ++i)
                {
                    pointCoord(0) = points[i][0];
                    pointCoord(1) = points[i][1];
                    // locate position in the initial grid
                    int ipos = static_cast<int>(pointCoord(0)) +  static_cast<int>(pointCoord(1)) * (nbMeshX + 1);
                    // reconstruct value from cuts coefficients
                    double valReconst = (*pairMA.second)(0, i) + (*pairMA.second)(1, i) * pointCoord(0) + (*pairMA.second)(2, i) * pointCoord(1);
                    BOOST_CHECK_CLOSE(valReconst, regressedValues(ipos, isim), accuracyEqual);
                }
            }
        }

        // default non compression
        BinaryFileArchive ar("archiveLLR2", "w");
        ar << Record(contCut, "FirstContinuation", "Top") ;

    }
    {
        // read archive
        BinaryFileArchive ar("archiveLLR2", "r");
        ContinuationCutsGridAdaptNonConcave contCut;
        Reference< ContinuationCutsGridAdaptNonConcave >(ar, "FirstContinuation", "Top").restore(0, &contCut);

        // calculate all cuts
        hyperStock(0, 0) = xMin;
        hyperStock(0, 1) = xMax;
        hyperStock(1, 0) = yMin;
        hyperStock(1, 1) = yMax;

        // test only values
        ArrayXi   pointNumber(4);
        ArrayXd pointCoord(2);
        // point test
        ArrayXd pointCoordTest(2);
        ArrayXd regPoint = ArrayXd::Constant(p_nDim, 0.5) ;
        for (int isim = 0; isim < p_nbSimul;  ++isim)
        {
            vector< pair<shared_ptr< GridAdaptBase>, shared_ptr<ArrayXXd>  > >   regressedCutsPMesh = contCut.getCutsASim(hyperStock, x.col(isim));

            for (const auto &pairMA : regressedCutsPMesh)
            {
                // for all points on the sub grid
                std::vector< Eigen::ArrayXd > points = pairMA.first->getPoints();

                for (size_t i = 0; i < points.size(); ++i)
                {
                    pointCoord(0) = points[i][0];
                    pointCoord(1) = points[i][1];
                    // reconstruct value from cuts coefficients : just to test
                    double valReconst = (*pairMA.second)(0, i) + (*pairMA.second)(1, i) * pointCoord(0) + (*pairMA.second)(2, i) * pointCoord(1);
                    BOOST_CHECK(abs(valReconst) < 1e10);
                }
            }
        }
    }
}



BOOST_AUTO_TEST_CASE(testContCutGridAdapt2D)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif

    TestContinuationCutsGridAdapt2D(1, 10, 1);
}


BOOST_AUTO_TEST_CASE(testCloneLocal)
{
#if defined   __linux
    enable_abort_on_floating_point_exception();
#endif
    // create the mesh
    ArrayXi  nbMesh = ArrayXi::Constant(2, 2);
    // generator Mersene Twister
    boost::mt19937 generator;
    boost::normal_distribution<double> alea_n;
    boost::variate_generator<boost::mt19937 &, boost::normal_distribution<double> > normal_random(generator, alea_n) ;
    std::function< double(double) > f = Func1();
    ArrayXXd x = ArrayXXd::Random(2, 100);
    // second member to regress
    ArrayXd toRegress(100);
    ArrayXd toReal(100);
    for (int is  = 0;  is < 100; ++is)
    {
        double prod = f(x(0, is));
        for (int id = 1 ; id < 2 ; ++id)
            prod *=  f(x(id, is));
        toRegress(is) = prod +  4 * normal_random();
        toReal(is) = prod;
    }
    // check if regression in t=0
    {
        // constructor if time t  = 0
        bool bZeroDate = true;
        // constructor
        LocalLinearRegression localRegressor(nbMesh);
        localRegressor.updateSimulations(bZeroDate, x);
        ArrayXd  regressedValues = localRegressor.getAllSimulations(toRegress);

        // now clone
        shared_ptr< BaseRegression> cloneRegressor = localRegressor.clone();
        ArrayXd  regressedValuesClone = cloneRegressor->getAllSimulations(toRegress);

        for (int is = 0; is < 100; ++is)
            BOOST_CHECK_CLOSE(regressedValues(is), regressedValuesClone(is), accuracyEqual);
    }

    {
        // constructor not zero
        bool bZeroDate = 0;

        // constructor
        LocalLinearRegression localRegressor(nbMesh);
        localRegressor.updateSimulations(bZeroDate, x);
        // first regression
        ArrayXd  regressedValues = localRegressor.getAllSimulations(toRegress);
        // now clone
        shared_ptr< BaseRegression> cloneRegressor = localRegressor.clone();
        ArrayXd  regressedValuesClone = cloneRegressor->getAllSimulations(toRegress);

        for (int is = 0; is < 100; ++is)
            BOOST_CHECK_CLOSE(regressedValues(is), regressedValuesClone(is), accuracyEqual);


    }
}
